#include "ambenv_view.h"

#include "GLCam.h"
#include "Instance.h"
#include "Mesh.h"
#include "util.h"

#include "ambenv_sdl.h"

using namespace ambenv;


// -- Variables global to this file.
const GLenum AE_COLOR_TEXTURE = GL_TEXTURE0;
const GLenum AE_CUBE_TEXTURE = GL_TEXTURE1;

Camera g_cameraPersp( "viewCam" );
GLCam g_cameraView( Vec3( 0.0, 100.0, -100.0 ), Vec3(), Vec3( 0.0, 1.0, 0.0 ) );


// -- Object rotation and rotational velocity.
GLfloat g_objectVelocity = 1.0;

GLfloat g_objectRotX = 0.0;
GLfloat g_objectRotZ = 0.0;

GLfloat g_objectVelX = 0.0;
GLfloat g_objectVelZ = 0.0;


GLuint g_displayList = 0;
GLuint g_sphereList = 0;
GLuint g_cubeBase = 0;

unsigned g_mode = AE_DRAW_FLAT;
unsigned g_cubeMap = 0;


// -- Function prototypes
int Init( Scene & scene );
int HandleEvent( SDL_Event * event );

void DrawAxes( GLfloat scale );
void DrawScene( Scene & scene );

bool HaveGLExtension( const std::string & extName );
void LoadTextures( Scene & scene );
void LoadCubeMap( unsigned cubeMapId );
void DeleteTextures( Scene & scene );
void DeleteDisplayList();



// -- Main drawing loop after initialization.
int ambenv::view( Scene & scene ) 
{
  
  // -- Initialize SDL video and OpenGL context.
  if( Init( scene ) != 0 ) {
    return( 1 );
  }
  
  // -- Load all textures in scene.textures into video memory.
  LoadTextures( scene );

  // -- Set up the perspective cameras.
  g_cameraPersp.perspective( 60, (GLfloat)scene.width / (GLfloat)scene.height, 1.0, 5000.0 );
  
  
  glClearColor( 0.0, 0.0, 0.0, 0.0 );
  glEnable( GL_DEPTH_TEST );
  glDepthFunc( GL_LEQUAL );
  glPolygonOffset( 1.0, 1.0 );
  
  glDisable( GL_NORMALIZE );
  glDisable( GL_CULL_FACE );
  
  glFrontFace( GL_CCW );
  
  
  if( g_sphereList == 0 ) {
    g_sphereList = glGenLists( 1 );
    glNewList( g_sphereList, GL_COMPILE );

    glActiveTexture( AE_CUBE_TEXTURE );

    glEnable( GL_TEXTURE_CUBE_MAP );

    glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP );
    glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP );
    glTexGeni( GL_R, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP );

    glEnable( GL_TEXTURE_GEN_S );
    glEnable( GL_TEXTURE_GEN_T );
    glEnable( GL_TEXTURE_GEN_R );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glPolygonMode( GL_FRONT, GL_LINE );
    glPolygonMode( GL_BACK, GL_FILL );
    glColor3f( 1.0, 1.0, 1.0 );
    GLUquadric * q = gluNewQuadric();
    gluSphere( q, 1000, 20, 20 );
    gluDeleteQuadric( q );
    
    
    glDisable( GL_TEXTURE_GEN_S );
    glDisable( GL_TEXTURE_GEN_T );
    glDisable( GL_TEXTURE_GEN_R );

    glActiveTexture( AE_COLOR_TEXTURE );

    glEndList();
  }


  /* Loop until done. */
  int done = 0;
  int frames = 0;
  Uint32 start_time = SDL_GetTicks();
  while( !done ) {

    g_objectRotX += g_objectVelX;
    g_objectRotZ += g_objectVelZ;

    if( g_objectRotX >= 360.0 ) {
      g_objectRotX -= 360.0;
    }
    if( g_objectRotZ >= 360.0 ) {
      g_objectRotZ -= 360.0;
    }
    if( g_objectRotX < 0.0 ) {
      g_objectRotX += 360.0;
    }
    if( g_objectRotZ < 360.0 ) {
      g_objectRotZ += 360.0;
    }


    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
    if( g_displayList == 0 ) {
      g_displayList = glGenLists( 1 );
      glNewList( g_displayList, GL_COMPILE );

      DrawScene( scene );

      glEndList();
    }


    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    doGLMultMatrix( g_cameraPersp.getPersp() );
    g_cameraView.load();
    
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    if( g_mode & AE_DRAW_CUBE_MAP ) {
      glCallList( g_sphereList );
    }

    glRotatef( g_objectRotX, 1, 0, 0 );
    glRotatef( g_objectRotZ, 0, 0, 1 );


    glCallList( g_displayList );

    glFinish();



    SDL_GL_SwapBuffers( );

    /* Check for error conditions. */
    GLenum gl_error = glGetError();
    if( gl_error != GL_NO_ERROR ) {
      scene.log.lm_fprintf( lm_err, "ambenv_view: OpenGL error (%d).\n", (int)gl_error );
    }

    char * sdl_error = SDL_GetError( );

    if( sdl_error[0] != '\0' ) {
      scene.log.lm_fprintf( lm_err, "ambenv_view: SDL error (%s).\n", sdl_error );
      SDL_ClearError();
    }

    /* Check if there's a pending event. */
    SDL_Event event;
    while( SDL_PollEvent( &event ) ) {
      done = HandleEvent( &event );
    }

    frames++;
  }
  Uint32 end_time = SDL_GetTicks();
  
  scene.log.lm_fprintf( lm_info, "Averaged %2.2f FPS\n",
    ((float)frames/(end_time-start_time))*1000.0 );
  	
  DeleteDisplayList();
  DeleteTextures( scene );

  
  /* Destroy our GL context, etc. */
  SDL_Quit();
  return( 0 );
}

void DrawAxes( GLfloat scale )
{
  GLfloat offplane = 0.0f;
  glColor3ub( 255, 64, 64 );
  glBegin( GL_LINE_STRIP );
  glVertex3f( 0.0f, offplane, 0.0f );
  glVertex3f( scale, offplane, 0.0f );
  glEnd();
  glColor3ub( 150, 255, 150 );
  glBegin( GL_LINE_STRIP );
  glVertex3f( 0.0f, offplane, 0.0f );
  glVertex3f( 0.0f, scale, 0.0f );
  glEnd();
  glColor3ub( 128, 128, 255 );
  glBegin( GL_LINE_STRIP );
  glVertex3f( 0.0f, offplane, 0.0f );
  glVertex3f( 0.0f, offplane, scale );
  glEnd();
}

// -- Taken from the Redbook.
bool HaveGLExtension( const std::string & extName )
{
  char * p = (char *)glGetString( GL_EXTENSIONS );
  
  char * end;
  if( p == NULL ) {
    return( false );
  }
  end = p + strlen( p );
  
  while( p < end ) {
    int n = strcspn( p, " " );
    if( ((int)extName.size() == n) && (strncmp( extName.c_str(), p, n ) == 0 ) ) {
      return( true );
    }
    p += (n + 1);
  }
  
  return( false );
}

int Init( Scene & scene )
{
  scene.bpp = 0;
  int rgb_size[3];
  int noframe = 0;
  Uint32 video_flags;
  int value = 0;
  
  if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
    scene.log.lm_fprintf( lm_err, "Could not initialize SDL: %s\n", SDL_GetError() );
    return( 1 );
  }
  
  const SDL_VideoInfo* info = SDL_GetVideoInfo();
  if( info == NULL ) {
    scene.log.lm_fprintf( lm_err, "SDL video query failed: %s\n", SDL_GetError() );
    return( 1 );
  }
  
  /* See if we should detect the display depth */
  if( scene.bpp == 0 ) {
#ifdef __USING_X11
    if( info->vfmt->BitsPerPixel <= 8 ) {
      scene.bpp = 8;
    } else {
      scene.bpp = 16;
    }
#else
    scene.bpp = info->vfmt->BitsPerPixel; 
#endif
  }
  
  video_flags = SDL_OPENGL;
  if( noframe ) {
    video_flags |= SDL_NOFRAME;
  }
  if( scene.fullscreen ) {
    video_flags |= SDL_FULLSCREEN;
  }
  
  /* Initialize the display */
  switch( scene.bpp ) {
  case 8:
    rgb_size[0] = 3;
    rgb_size[1] = 3;
    rgb_size[2] = 2;
    break;
  case 15:
  case 16:
    rgb_size[0] = 5;
    rgb_size[1] = 5;
    rgb_size[2] = 5;
    break;
  default:
    rgb_size[0] = 8;
    rgb_size[1] = 8;
    rgb_size[2] = 8;
    break;
  }
  
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, rgb_size[0] );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, rgb_size[1] );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, rgb_size[2] );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, scene.bpp );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  if ( SDL_SetVideoMode( scene.width, scene.height, scene.bpp, video_flags ) == NULL ) {
    scene.log.lm_fprintf( lm_err, "Could not set GL mode: %s\n", SDL_GetError() );
    SDL_Quit();
    return( 1 );
  }
  
  scene.log.lm_fprintf( lm_info, "Screen BPP: %d\n\n", 
    SDL_GetVideoSurface()->format->BitsPerPixel );
  scene.log.lm_fprintf( lm_info, "Vendor    : %s\n", glGetString( GL_VENDOR ) );
  scene.log.lm_fprintf( lm_info, "Renderer  : %s\n", glGetString( GL_RENDERER ) );
  scene.log.lm_fprintf( lm_info, "Version   : %s\n", glGetString( GL_VERSION ) );
  scene.log.lm_fprintf( lm_info, "Extensions: %s\n\n", glGetString( GL_EXTENSIONS ) );
  
  SDL_GL_GetAttribute( SDL_GL_RED_SIZE, &value );
  scene.log.lm_fprintf( lm_info, "SDL_GL_RED_SIZE:     requested %d, got %d\n", 
    rgb_size[0], value );
  SDL_GL_GetAttribute( SDL_GL_GREEN_SIZE, &value );
  scene.log.lm_fprintf( lm_info, "SDL_GL_GREEN_SIZE:   requested %d, got %d\n", 
    rgb_size[1], value );
  SDL_GL_GetAttribute( SDL_GL_BLUE_SIZE, &value );
  scene.log.lm_fprintf( lm_info, "SDL_GL_BLUE_SIZE:    requested %d, got %d\n", 
    rgb_size[2], value );
  SDL_GL_GetAttribute( SDL_GL_DEPTH_SIZE, &value );
  scene.log.lm_fprintf( lm_info, "SDL_GL_DEPTH_SIZE:   requested %d, got %d\n", 
    scene.bpp, value );
  SDL_GL_GetAttribute( SDL_GL_DOUBLEBUFFER, &value );
  scene.log.lm_fprintf( lm_info, "SDL_GL_DOUBLEBUFFER: requested 1, got %d\n\n", 
    value );
  
  /* Set the window manager title bar */
  SDL_WM_SetCaption( "ambenv_view", "ambenv_view" );
  

  // -- Make a check for the extensions we want to use.
  if( !HaveGLExtension( "GL_ARB_texture_cube_map" ) && !HaveGLExtension( "GL_EXT_texture_cube_map" ) ) {
    scene.log.lm_fprintf( lm_err, "Requires missing cube map extension: %s, %s\n", "GL_ARB_texture_cube_map", "GL_EXT_texture_cube_map" );
  }
  if( !HaveGLExtension( "GL_ARB_multitexture" ) ) {
    scene.log.lm_fprintf( lm_info, "Requires missing multi-texturing extension: %s\n", "GL_ARB_multitexture" );
  }


#ifdef WIN32
  glActiveTexture = (PFNGLACTIVETEXTUREPROC)wglGetProcAddress( "glActiveTextureARB" );
  glMultiTexCoord2f = (PFNGLMULTITEXCOORD2FPROC)wglGetProcAddress( "glMultiTexCoord2fARB" );
  assert( glActiveTexture );
  assert( glMultiTexCoord2f );
#endif

  return( 0 );
}

void DeleteDisplayList()
{
  if( g_displayList ) {
    glDeleteLists( g_displayList, 1 );
    g_displayList = 0;
  }
}

void DeleteTextures( Scene & scene )
{
  if( scene.textures.size() > 0 ) {
    GLuint textureBase = (scene.textures.begin()->second)->m_glTextureId;
    GLuint numTextures = scene.textures.size();
    glDeleteTextures( numTextures, &textureBase );
  }
}

void DrawScene( Scene & scene )
{
  // -- Create a map from textureId to Texture.
  std::vector<Texture *> tid_map( scene.textures.size() + 1 );
  {
    std::map<std::string, Texture *>::iterator t_itr = scene.textures.begin();
    for( ; t_itr != scene.textures.end(); ++t_itr ) {
      assert( t_itr->second->m_textureId < tid_map.size() );
      
      tid_map[t_itr->second->m_textureId] = t_itr->second;
    }
  }
   
  // -- Create a map from materialId to Material.
  std::vector<Material *> mid_map( scene.materials.size() + 1 );
  {
    unsigned i=0;
    std::map<std::string, Material *>::iterator m_itr = scene.materials.begin();
    for( ; m_itr != scene.materials.end(); ++m_itr, i++ ) {
      assert( m_itr->second->materialId() < mid_map.size() );
      
      mid_map[m_itr->second->materialId()] = m_itr->second;
    }
  }
 

  glActiveTexture( AE_CUBE_TEXTURE );
  if( g_mode & AE_DRAW_CUBE_MAP ) {    
    glEnable( GL_TEXTURE_CUBE_MAP );

    glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP );
    glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP );
    glTexGeni( GL_R, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP );

    glEnable( GL_TEXTURE_GEN_S );
    glEnable( GL_TEXTURE_GEN_T );
    glEnable( GL_TEXTURE_GEN_R );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
  } else {
    glDisable( GL_TEXTURE_CUBE_MAP );
  }


  glActiveTexture( AE_COLOR_TEXTURE );
  if( g_mode & AE_DRAW_TEXTURE ) {
    glEnable( GL_TEXTURE_2D );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
  } else {
    glDisable( GL_TEXTURE_2D );
  }

  glDisable( GL_CULL_FACE );


  // -- Setup some default parameters.
  glPolygonMode( GL_BACK, GL_LINE );
  glPolygonMode( GL_FRONT, GL_FILL );
  glShadeModel( GL_SMOOTH );

  
  GLuint currentTexture = 0;
  std::list<Instance *>::const_iterator itr = scene.instances.begin();
  for( ; itr != scene.instances.end(); ++itr ) {
    
    const Mesh * mesh = (Mesh *)((*itr)->object());
    assert( mesh != NULL );
    
    glPushMatrix();
    doGLMultMatrix( (*itr)->getMatrix() );
    
    Vec3 tcScale( 1.0, 1.0, 1.0 );    
    
    glBegin( GL_TRIANGLES );
    
    for( unsigned i=0; i<mesh->numTriangles(); i++ ) {
      
      MeshTriangle tri = mesh->getTriangle( i );	 
      
      if( tri.textureId() > 0 ) {
        // -- This triangle is textured.
        
        assert( tri.textureId() < tid_map.size() );
        
        // -- If this triangle's texture is not bound already, bind it.
        Texture * tp = tid_map[tri.textureId()];
        if( tp->m_glTextureId != currentTexture ) {
          glEnd();
          glBindTexture( GL_TEXTURE_2D, tp->m_glTextureId );
          currentTexture = tp->m_glTextureId;
          tcScale.x( tp->m_tcoord[2] );
          tcScale.y( tp->m_tcoord[3] );
          glBegin( GL_TRIANGLES );
        }
        
        // -- Get the underlying diffuse surface color, defaulting to 1.
        Vec3 diffuse( 1.0, 1.0, 1.0 );
        if( tri.materialId() > 0 ) {
          diffuse = mid_map[tri.materialId()]->diffuse();
        }
        
        // -- Output vertices.
        for( unsigned j=0; j<3; j++ ) {
          MeshVertex vertex = mesh->getVertex( tri.vert( j ) );
          Vec3 v = vertex.v();
          Vec3 n;
          
          Vec3 tc = vertex.tc() * tcScale;
          
          // -- Should we use occlusion values?
          if( g_mode & AE_DRAW_PVO ) {
            Vec3 d_occl = diffuse * vertex.occl();
            glColor3f( d_occl.x(), d_occl.y(), d_occl.z() );
	  } else {
	    glColor3f( diffuse.x(), diffuse.y(), diffuse.z() );
	  }
	  if( g_mode & AE_DRAW_BENT_NORMAL ) {
            n = vertex.bentNormal();
          } else {
            n = vertex.n();
          }
          
          glMultiTexCoord2f( AE_COLOR_TEXTURE, tc.x(), tc.y() );
          glNormal3f( n.x(), n.y(), n.z() );
          glVertex3f( v.x(), v.y(), v.z() );
        }
        
      } else {
        // -- This vertex does not have a color texture map.
        // -- Fall back to vertex material color.
        
        glEnd();

	if( g_mode & AE_DRAW_TEXTURE ) {
	  glDisable( GL_TEXTURE_2D );
	}
        
        
        Vec3 diffuse( 1.0, 1.0, 1.0 );
        if( tri.materialId() > 0 ) {
          diffuse = mid_map[tri.materialId()]->diffuse();
        }
        glColor3f( diffuse.x(), diffuse.y(), diffuse.z() );
        

        glBegin( GL_TRIANGLES );
        for( unsigned j=0; j<3; j++ ) {

          MeshVertex vertex = mesh->getVertex( tri.vert( j ) );
          Vec3 v = vertex.v();
          Vec3 n;
          if( g_mode & AE_DRAW_PVO ) {
            Vec3 d_occl = diffuse * vertex.occl();
            glColor3f( d_occl.x(), d_occl.y(), d_occl.z() );
	  }
	  if( g_mode & AE_DRAW_BENT_NORMAL ) {
            n = vertex.bentNormal();
          } else {
            n = vertex.n();
          }
          glNormal3f( n.x(), n.y(), n.z() );
          glVertex3f( v.x(), v.y(), v.z() );
        }
        glEnd();

	if( g_mode & AE_DRAW_TEXTURE ) {
	  glEnable( GL_TEXTURE_2D );
	}
        glBegin( GL_TRIANGLES );
        
      }
    }
    glEnd();

    
    if( g_mode & AE_DRAW_NORMALS ) {
      for( unsigned i=0; i<mesh->numVertices(); i++ ) {
        Vec3 v = mesh->getVertex( i ).v();
        Vec3 n = mesh->getVertex( i ).n();
        Vec3 bent = mesh->getVertex( i ).bentNormal();
        
        glColor3f( 1.0, 1.0, 0.0 );
        
        n = v + n*2.0;
        
        glBegin( GL_LINE_STRIP );
        glVertex3f( v.x(), v.y(), v.z() );
        glVertex3f( n.x(), n.y(), n.z() );
        glEnd();
        
        glColor3f( 1.0, 0.0, 0.0 );
        n = v + bent * 2.0;
        glBegin( GL_LINE_STRIP );
        glVertex3f( v.x(), v.y(), v.z() );
        glVertex3f( n.x(), n.y(), n.z() );
        glEnd();
      }
    }

    if( g_mode & AE_DRAW_EDGES ) {
      glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
      glEnable( GL_POLYGON_OFFSET_FILL );
      glColor3f( 0.4, 0.4, 0.4 );
      for( unsigned i=0; i<mesh->numTriangles(); i++ ) {
	MeshTriangle tri = mesh->getTriangle( i );
	glBegin( GL_TRIANGLES );
        for( unsigned j=0; j<3; j++ ) {
          Vec3 v = mesh->getVertex( tri.vert( j ) ).v();
          glVertex3f( v.x(), v.y(), v.z() );
        }
        glEnd();
      }
      glPolygonMode( GL_BACK, GL_LINE );
      glPolygonMode( GL_FRONT, GL_FILL );
      glDisable( GL_POLYGON_OFFSET_FILL );
    }
 
    glPopMatrix();
  }

  

  if( g_mode & AE_DRAW_TEXTURE ) {
    glDisable( GL_TEXTURE_2D );
  }

  if( g_mode & AE_DRAW_CUBE_MAP ) {
    
    glActiveTexture( AE_CUBE_TEXTURE );
    
    glDisable( GL_TEXTURE_GEN_S );
    glDisable( GL_TEXTURE_GEN_T );
    glDisable( GL_TEXTURE_GEN_R );
    
    glDisable( GL_TEXTURE_CUBE_MAP );
  }

  glActiveTexture( AE_COLOR_TEXTURE );

}

int HandleEvent( SDL_Event * event )
{
  int done;
  static int startx = 0;
  static int starty = 0;
  static int shift = 0;
  static int ctrl = 0;
  Uint8 mstate = 0;
  
  done = 0;
  switch( event->type ) {
  case SDL_ACTIVEEVENT:
    break;
  case SDL_MOUSEBUTTONDOWN:
    mstate = SDL_GetMouseState( &startx, &starty );
    break;
  case SDL_MOUSEBUTTONUP:
    break;
  case SDL_MOUSEMOTION:
    {
      int currentx = 0;
      int currenty = 0;
      mstate = SDL_GetMouseState( &currentx, &currenty );
      
      // -- moving the regular camera
      if( mstate & SDL_BUTTON(1) ) {
        if( shift ) {
          g_cameraView.orbitXZ( (Real)(currentx - startx) );
          g_cameraView.orbitXY( (Real)(currenty - starty) );
        } else if( ctrl ) {
          g_cameraView.dolly( (Real)(currenty - starty) );
        } else {
          g_cameraView.translate( (Real)(currentx - startx), (Real)(currenty - starty) );
        }
      }
      
      startx = currentx;
      starty = currenty;
    }
    break;
  case SDL_KEYDOWN:
    if( event->key.keysym.sym == SDLK_ESCAPE ) {
      done = 1;
    } else if( (event->key.keysym.sym == SDLK_LSHIFT) || 
	       (event->key.keysym.sym == SDLK_RSHIFT) ) {
      shift = 1;
    } else if( (event->key.keysym.sym == SDLK_LCTRL) || 
	       (event->key.keysym.sym == SDLK_RCTRL) ) {
      ctrl = 1;
    } else if( event->key.keysym.sym == SDLK_UP ) {
      g_objectVelZ = g_objectVelocity;
    } else if( event->key.keysym.sym == SDLK_DOWN ) {
      g_objectVelZ = -g_objectVelocity;
    } else if( event->key.keysym.sym == SDLK_RIGHT ) {
      g_objectVelX = g_objectVelocity;
    } else if( event->key.keysym.sym == SDLK_LEFT ) {
      g_objectVelX = -g_objectVelocity;
    } else if( event->key.keysym.sym == SDLK_1 ) {
      g_mode = AE_DRAW_FLAT;
      DeleteDisplayList();
    } else if( event->key.keysym.sym == SDLK_2 ) {
      g_mode ^= AE_DRAW_TEXTURE;
      DeleteDisplayList();
    } else if( event->key.keysym.sym == SDLK_3 ) {
      g_mode ^= AE_DRAW_CUBE_MAP;
      DeleteDisplayList();
    } else if( event->key.keysym.sym == SDLK_4 ) {
      g_mode ^= AE_DRAW_PVO;
      DeleteDisplayList();
    } else if( event->key.keysym.sym == SDLK_5 ) {
      g_mode ^= AE_DRAW_BENT_NORMAL;
      DeleteDisplayList();
    } else if( event->key.keysym.sym == SDLK_6 ) {
      g_mode ^= AE_DRAW_NORMALS;
      DeleteDisplayList();
    } else if( event->key.keysym.sym == SDLK_7 ) {
      g_mode ^= AE_DRAW_EDGES;
      DeleteDisplayList();
    } else if( event->key.keysym.sym == SDLK_q ) {
      if( g_cubeMap != 1 ) {
        LoadCubeMap( 1 );
      }
    } else if( event->key.keysym.sym == SDLK_w ) {
      if( g_cubeMap != 2 ) {
        LoadCubeMap( 2 );
      }
    } else if( event->key.keysym.sym == SDLK_e ) {
      if( g_cubeMap != 3 ) {
        LoadCubeMap( 3 );
      }
    } else if( event->key.keysym.sym == SDLK_r ) {
      if( g_cubeMap != 4 ) {
        LoadCubeMap( 4 );
      }
    } else if( event->key.keysym.sym == SDLK_t ) {
      if( g_cubeMap != 5 ) {
        LoadCubeMap( 5 );
      }
    } else if( event->key.keysym.sym == SDLK_PLUS ) {
      g_objectVelocity *= 2.0;
    } else if( event->key.keysym.sym == SDLK_MINUS ) {
      g_objectVelocity *= 0.5;
    }
    break;
  case SDL_KEYUP:
    if( event->key.keysym.sym == SDLK_LSHIFT ) {
      shift = 0;
    } else if( event->key.keysym.sym == SDLK_LCTRL ) {
      ctrl = 0;
    } else if( event->key.keysym.sym == SDLK_UP ) {
      if( g_objectVelZ > 0 ) {
        g_objectVelZ = 0;
      }
    } else if( event->key.keysym.sym == SDLK_DOWN ) {
      if( g_objectVelZ < 0 ) {
        g_objectVelZ = 0;
      }
    } else if( event->key.keysym.sym == SDLK_RIGHT ) {
      if( g_objectVelX > 0 ) {
        g_objectVelX = 0;
      }
    } else if( event->key.keysym.sym == SDLK_LEFT ) {
      if( g_objectVelX < 0 ) {
        g_objectVelX = 0;
      }
    }
    break;
  case SDL_QUIT:
    done = 1;
    break;
  }
  
  return( done );
}

void LoadTextures( Scene & scene )
{
  int totalBytes = 0;
  GLuint textureBase = 0;
  GLuint textureOffset = 0;
  
  glGenTextures( scene.textures.size(), &textureBase );
  
  std::map<std::string, Texture *>::iterator itr = scene.textures.begin();
  for( ; itr != scene.textures.end(); ++itr ) {
    Texture * pt = itr->second;
    
    unsigned width = power_of_two( pt->m_width );
    unsigned height = power_of_two( pt->m_height );
    
    pt->m_tcoord[0] = 0.0f;
    pt->m_tcoord[1] = 0.0f;
    pt->m_tcoord[2] = (GLfloat)pt->m_width / (GLfloat)width;
    pt->m_tcoord[3] = (GLfloat)pt->m_height / (GLfloat)height;
    
    std::vector<unsigned char> padded( 4*width*height );
    for( unsigned i=0; i<pt->m_height; i++ ) {
      for( unsigned j=0; j<pt->m_width; j++ ) {
        for( unsigned k=0; k<4; k++ ) {
          padded[4*(width*i + j) + k] = pt->m_data[4*(pt->m_width*i + j) + k];
        }
      }
    }
    
    pt->m_glTextureId = (textureBase + textureOffset);
    
    glBindTexture( GL_TEXTURE_2D, pt->m_glTextureId );
    
    gluBuild2DMipmaps( GL_TEXTURE_2D, GL_RGBA8, width, height,
      GL_RGBA, GL_UNSIGNED_BYTE, &padded[0] );
    //glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, 
    //		  GL_RGBA, GL_UNSIGNED_BYTE, &padded[0] );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
      GL_NEAREST_MIPMAP_LINEAR );
    
    
    totalBytes += padded.size();
    textureOffset++;
  }
  
  LoadCubeMap( 1 );
  
  // -- Check for GL errors, just for fun. 
  GLenum gl_error = glGetError();
  if( gl_error != GL_NO_ERROR ) {
    scene.log.lm_fprintf( lm_err, "OpenGL error code (%d) during texture load.\n", 
      (int)gl_error );
  }
  
  scene.log.lm_fprintf( lm_info, "Total %.2f MB textures loaded.\n\n", 
    (float)totalBytes / 1048576.0 );
}

void LoadCubeMap( unsigned cubeMapId )
{
  if( (cubeMapId > NUM_CUBE_MAPS) || (cubeMapId <= 0) ) {
    return;
  }
  
  g_cubeMap = cubeMapId;
  
  glActiveTexture( AE_CUBE_TEXTURE );

  if( g_cubeBase ) {
    glDeleteTextures( 1, &g_cubeBase );
  }
  glGenTextures( 1, &g_cubeBase );
  glBindTexture( GL_TEXTURE_CUBE_MAP, g_cubeBase );

  for( unsigned i=0; i<6; i++ ) {
    unsigned width = 0, height = 0;
    std::vector<unsigned char> data;
    
    read_rgba_tiff( g_cube_filename[g_cubeMap-1][i], width, height, data );
    
    glTexImage2D( g_cube_face[i], 0, GL_RGBA, width, height,
      0, GL_RGBA, GL_UNSIGNED_BYTE, &data[0] );
  }
  
  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
}
