#include <SDL/SDL.h>

#include "ambenv_view.h"
#include "AmbOccl.h"
#include "util.h"

#include <queue>
#include <string>
#include <fstream>

using namespace ambenv;

bool ParseCommandLine( int argc, char * argv[], Scene & scene );

int main( int argc, char * argv[] )
{
  Scene scene;
  scene.samples = 64;

  if( !ParseCommandLine( argc, argv, scene ) ) {
    scene.log.lm_fprintf( lm_out, "Usage: %s [-whfsh] mesh_files\n", argv[0] );
    return( 0 );
  }


  // -- Open log files.
  std::ofstream info( "info.log", std::ofstream::out );
  if( info.is_open() ) {
    info << "# -- info.log from (" << argv[0] << ")\n\n";
    scene.log.setInfo( info );
  }
  std::ofstream warn( "warn.log", std::ofstream::out );
  if( warn.is_open() ) {
    warn << "# -- warn.log from (" << argv[0] << ")\n\n";
    scene.log.setWarn( warn );
  }

  for( unsigned i=0; i<scene.input_files.size(); i++ ) {
    if( !scene.load3DS( scene.input_files[i] ) ) {
      scene.log.lm_fprintf( lm_err, "Error loading file (%s)\n", 
			    scene.input_files[i].c_str() );
    }
  }

  std::cout << scene << "\n";

  if( scene.usehemicube ) {
    generate_pv_hemicube( scene );
  } else {
    generate_pv( scene );
  }
 
  view( scene );

  
  // -- Close log files.
  info << "# -- EOF\n";
  info.close();
  warn << "# -- EOF\n";
  warn.close();

  return( 0 );
}

const int NUM_OPTIONS = 10;
const std::string optionString[NUM_OPTIONS] = {
  "--width",
  "-w",
  "--height",
  "-h",
  "--samples",
  "-s",
  "--fullscreen",
  "-f",
  "--hemicube",
  "-c"
};

const int optionNumArgs[NUM_OPTIONS] = { 1, 1, 1, 1, 1, 1, 0, 0, 0, 0 };

bool ParseCommandLine( int argc, char * argv[], Scene & scene )
{
  if( argc < 2 ) {
    return( false );
  }

  scene.outfile = "";

  std::vector<std::string> input_files;
  std::queue<std::string> q;
  {
  // -- Separate the options and arguments from the input files.
  for( int i=1; i<argc; i++ ) {
    bool isoption = false;
    for( int j=0; j<NUM_OPTIONS; j++ ) {
      if( optionString[j] == argv[i] ) {
	q.push( argv[i] );
	int numArgs = optionNumArgs[j];
	if( i + numArgs >= argc ) {
	  return( false );
	}
	for( int k=0; k<numArgs; k++ ) {
	  i++;
	  q.push( argv[i] );
	}

	isoption = true;
	break;
      }
    }
    if( !isoption && i < argc ) {
      input_files.push_back( argv[i] );
    }
  }
  }
  if( input_files.empty() ) {
    return( false );
  }


  // -- Validate and save the arguements.
  while( !q.empty() ) {
    if( ("--width" == q.front()) || ("-w" == q.front()) ) {
      q.pop();
      if( q.empty() ) {
	return( false );
      }
      int val = (int)strtol( q.front().c_str(), (char **)NULL, 10 );
      if( (ERANGE == errno) || (0 >= val) ) {
	std::cerr << argv[0] << ": Bad width value (" << q.front() << ")\n";
	return( false );
      }
      scene.width = (unsigned)val;
      q.pop();
    } else if( ("--height" == q.front()) || ("-h" == q.front()) ) {
      q.pop();
      if( q.empty() ) {
        return( false );
      }
      int val = (int)strtol( q.front().c_str(), (char **)NULL, 10 );
      if( (ERANGE == errno) || (0 >= val) ) {
	std::cerr << argv[0] << ": Bad height value (" << q.front() << ")\n";
        return( false );
      }
      scene.height = (unsigned)val;
      q.pop();
    } else if( ("--fullscreen" == q.front()) || ("-f" == q.front()) ) {
      q.pop();
      scene.fullscreen = true;
    } else if( ("--hemicube" == q.front()) || ("-c" == q.front()) ) {
      q.pop();
      scene.usehemicube = true;
    } else if( ("--samples" == q.front()) || ("-s" == q.front()) ) {
      q.pop();
      if( q.empty() ) {
        return( false );
      }
      long val = strtol( q.front().c_str(), (char **)NULL, 10 );
      if( (ERANGE == errno) || (0 >= val) ) {
	std::cerr << argv[0] << ": Bad number of samples (" << q.front() << ")\n";
        return( false );
      }
      scene.samples = (unsigned)val;
      q.pop();
    } else {
      if( !q.empty() ) {
	scene.log.lm_fprintf( lm_err, "%s: Unknown option (%s)\n", 
			      argv[0], q.front().c_str() );
      }
      return( false );
    }
  }

  scene.input_files = input_files;
  {
  // -- Check if the input files exist.
  for( unsigned i=0; i<scene.input_files.size(); i++ ) {
    std::ifstream infile( scene.input_files[i].c_str() );
    if( !infile.is_open() ) {
      std::cerr << argv[0] << ": Could not read file (" << scene.input_files[i] << ")\n";
      return( false );
    }
    infile.close();
  }
  }
  return( true );
}
